#include "main.h"

#define F_SPI 400000
//maximum working frequency is 927000

struct spi_data
{
	union 
	{
		struct
		{
			unsigned char	u1_address	:3	;
			unsigned int	u1_data_a	:10	;
			unsigned int	u1_data_b	:10	;
			unsigned char	u1_watch_dog	:1	;
			unsigned char	u1_divisor	:4	;
			unsigned char	u1_dac_a	:4	;
			unsigned char	u1_dac_b	:4	;
	
			unsigned char	u2_address	:3	;
			unsigned int	u2_data_a	:10	;
			unsigned int	u2_data_b	:10	;
			unsigned char	u2_watch_dog	:1	;
			unsigned char	u2_divisor	:4	;
			unsigned char	u2_dac_a	:4	;
			unsigned char	u2_dac_b	:4	;
		} __attribute__ ((packed));
		unsigned char bytes[9];
	} __attribute__ ((packed));
} __attribute__ ((packed));

struct spi_status
{
	union 
	{
		struct
		{
			unsigned char	u1_status_a	:8	;
			unsigned char	u1_status_b	:8	;
			unsigned char	u1_watch_dog	:1	;
			unsigned char	u1_thermal	:1	;
			unsigned char	u1_revision	:3	;
			unsigned int	u1_filler	:15	;
	
			unsigned char	u2_status_a	:8	;
			unsigned char	u2_status_b	:8	;
			unsigned char	u2_watch_dog	:1	;
			unsigned char	u2_thermal	:1	;
			unsigned char	u2_revision	:3	;
			unsigned int	u2_filler	:15	;
		} __attribute__ ((packed));
		unsigned char bytes[9];
	} __attribute__ ((packed));
} __attribute__ ((packed));

void send_spi_msg(struct spi_data data)
{
	signed char i = 8;
	
	PORTC |= 0b10000000;
	SPCR = 0b01000000;
	SPSR = 0b00000000;
	if (SPSR & 0b10000000)
		SPDR = 0;
	SPDR = data.bytes[8];

	ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
	{
		while ((PINB & 0b10000000) != 0)
			;
		while ((PINB & 0b10000000) == 0)
			;
		while ((PINB & 0b10000000) != 0)
			;
		PORTC &= ~(0b10000000);
		do
		{
			i--;
			while ((SPSR & 0b10000000) == 0)
				;
			SPDR = data.bytes[i];
		}
		while (i > 0);
		_delay_us(8000000.0 / F_SPI);
		PORTC |= 0b10000000;
	}
}

unsigned char dac[4];
unsigned int pwm[4][8];

void initialize_leds(void)
{
	DDRB  |= 0b01001000;
	DDRB  &= 0b01001111;
	PORTB |= 0b00010000;
	PORTB &= 0b00010111;
	DDRC  |= 0b10000000;
	PORTC |= 0b10000000;
	
	SPCR   = 0b01000000;
	SPSR   = 0b00000000;
	
	OCR0 = (unsigned char)(0.5 + F_CPU / (2.0 * F_SPI));
	TCNT0 = 0;
	TCCR0 = 0b00011001;
	TIMSK = 0b00000000;

	set_multi_dac(1, 1, 1, 1);
	for (unsigned char i = 0; i < 8; i++)
		set_multi_pwm(i, 400, 400, 400, 400);
	_delay_ms(500);
	for (unsigned char i = 0; i < 8; i++)
		set_multi_pwm(i, 0, 0, 0, 0);
}

void update_leds(unsigned char address)
{
	if (address < 8)
	{
		struct spi_data data;
		
		data.u1_watch_dog = 1;
		data.u1_divisor = 0;
		data.u1_dac_a = dac[0];
		data.u1_dac_b = dac[1];
		data.u2_watch_dog = 1;
		data.u2_divisor = 0;
		data.u2_dac_a = dac[2];
		data.u2_dac_b = dac[3];
		
		data.u2_data_a = pwm[2][address];
		data.u2_data_b = pwm[3][address];
		data.u1_data_a = pwm[0][address];
		data.u1_data_b = pwm[1][address];
		
		data.u1_address = address;
		data.u2_address = address;
		
		send_spi_msg(data);
	}
}

void set_dac(unsigned char value, unsigned char bank)
{
	if (bank < 4)
		dac[bank] = value;
	update_leds(0);
}
void set_multi_dac(unsigned char value_0, unsigned char value_1, unsigned char value_2, unsigned char value_3)
{
	dac[0] = value_0;
	dac[1] = value_1;
	dac[2] = value_2;
	dac[3] = value_3;
	update_leds(0);
}
void set_pwm(unsigned char address, unsigned int value, unsigned char bank)
{
	if ((bank < 4) && (address < 8))
		pwm[bank][address] = value;
	update_leds(address);
}
void set_multi_pwm(unsigned char address, unsigned int value_0, unsigned int value_1, unsigned int value_2, unsigned int value_3)
{
	if (address < 8)
	{
		pwm[0][address] = value_0;
		pwm[1][address] = value_1;
		pwm[2][address] = value_2;
		pwm[3][address] = value_3;
	}
	update_leds(address);
}

