#include "main.h"


#define SEND_RING_CAPACITY 9
volatile char send_ring[SEND_RING_CAPACITY + 1];
volatile unsigned char send_ring_size = 0;
volatile unsigned char send_ring_offset = 0;

volatile struct recv_buffer recv_buf_a;
volatile struct recv_buffer recv_buf_b;
volatile struct recv_buffer * volatile recv_buf_full = &recv_buf_b;


void initialize_uart(void)
{
	DDRC  |= 0b00000100;
	PORTC &= 0b11111011;
	DDRD  |= 0b00000010;
	PORTD &= 0b11110000;

	UCSRA = 0b00000000;
	UCSRB = 0b00011000;
	UCSRC = 0b10001110;
	UBRRL = 25;
	UBRRH = 0;
	recv_buf_a.valid = 0;
	recv_buf_b.valid = 0;
	send_ring[SEND_RING_CAPACITY] = '+';
	UCSRB = 0b10111000;
}



ISR(USART_UDRE_vect)
{
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
	{
		if (send_ring[SEND_RING_CAPACITY] != '+')
		{
			while (1)
			{
				UDR = '6';
				_delay_ms(5);
				UDR = '\n';
				_delay_ms(995);
			}
		}
		if (send_ring_size > 0)
		{
			UDR = send_ring[send_ring_offset % SEND_RING_CAPACITY];
			send_ring_size--;
			send_ring_offset++;
			if (send_ring_offset >= SEND_RING_CAPACITY)
				send_ring_offset -= SEND_RING_CAPACITY;
		}
		else
		{
			UCSRB &= ~(0b00100000);
		}
	}
}
void send_char(char c)
{
	bool done = 0;
	while (!done)
	{
		ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
		{
			if (send_ring_size < SEND_RING_CAPACITY)
			{
				send_ring[(send_ring_size + send_ring_offset) % SEND_RING_CAPACITY] = c;
				send_ring_size++;
				UCSRB |= 0b00100000;
				done = 1;
			}
		}
	}
}

void send_int(signed int n1)
{
	unsigned int n2 = n1;
	if (n1 < 0)
	{
		send_char('-');
		n2 = -n1;
	}
	bool show = 0;
	const unsigned int factors[5] = {10000, 1000, 100, 10, 1};
	for (unsigned char i = 0; i < 5; i++)
	{
		unsigned int d = n2 / factors[i];
		n2 = n2 % factors[i];
		if ((d != 0) || (i == 4))
			show = 1;
		if (show == 1)
			send_char((char)d + '0');
	}
}

void send_response(signed int n)
{
	send_int(n);
	send_char('\n');
}


char recv_char_now(void)
{
	while ((UCSRA & 0b10000000) == 0)
		;
	char c = UDR;
	return c;
}


volatile struct recv_buffer * volatile curbuf = &recv_buf_a;

volatile struct recv_buffer * get_next_data(void)
{
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
	{
		if ((recv_buf_full->valid == 0) && (curbuf->valid != 0))
		{
			volatile struct recv_buffer * tmpbuf = recv_buf_full;
			recv_buf_full = curbuf;
			curbuf = tmpbuf;
			curbuf->size = 0;
		}
	}
	if (recv_buf_full->valid)
		return recv_buf_full;
	else
		return NULL;
}

ISR(USART_RX_vect)
{
	static bool overflow = 0;
	if (curbuf->valid)
	{
		get_next_data();
		if (curbuf->valid)
		{
			recv_char_now();
			overflow = 1;
			return;
		}
		else
		{
			if (overflow)
				curbuf->size = -1;
			else
				curbuf->size = 0;
			overflow = 0;
		}
	}
	unsigned char flags = UCSRA; //have to read it right now
	char c = recv_char_now();
	if ((flags & 0b00011100) != 0)
	{
		curbuf->size = -1;
		curbuf->valid = 1;
		return;
	}
	if (c == '\n')
	{
		curbuf->valid = 1;
		return;
	}
	if ((curbuf->size >= 23) || (curbuf->size < 0))
	{
		curbuf->size = -1;
	}
	else
	{
		curbuf->data[(unsigned char)curbuf->size] = c;
		curbuf->size++;
	}
	return;
	
}

